##########################################################################################
#This python script transforms binary character sequences from a BEAUti-generated BEAST 2
#  xml script to sequences with uncertainty coded in. The idea is to treat a 0 in the data
#  as the probability p that it is actually a zero, and (1-p) that it is actually 1 in the
#  ground truth. This transformation may be relevant for linguistic lexical data, where 
#  lexical characters may be recorded as 0s, or missing, just because the expert
#  literature is incomplete and fails to record an existing word. In line with the usual 
#  structure of linguistic lexical datasets, cognate presences (i.e. 1s) are treated as certain.
#The recommended range for P_THAT_ZERO_IS_ONE and P_THAT_?_IS_ONE arguments is [0-0.2].
#  Given the usual interpretation of 1s, 0s and ?s in linguistic lexical datasets (see the
#  note below for information on that), a value like 0.03 could be reasonable for 0s, and
#  a value like 0.1, reasonable for ?s. For more information, see case studies in Sec. 3 
#  of the corresponding paper.
#
#NOTE: The script presupposes conventional coding of linguistic lexical data wherein 1 
#  signifies the presence of a cognate in the given language, 0 its absence, and ? 
#  uncertainty, usually reserved for cases when existing expert literature does not 
#  record any cognate for a very basic meaning, such as e.g. "eye": it is very unlikely
#  that any language would miss a word with such meaning entirely. See Sec. 2 of the 
#  corresponding paper for details. 
#
##########################################################################################
#USAGE:
##########################################################################################
#0. Check that your python installation includes library xmltodict. (You can do that by 
#     starting the command-line python interface with command "python", and then 
#     entering command "import xmltodict" in the prompt. If this results in an error, you  
#     need to install xmltodict for your python first. 
# 
#1. Build a BEAST 2 xml for your linguistic data, using BEAUti.
#
#2. Copy the data node from your BEAST 2 xml into a separate text file. That node most 
#     likely is located very near the start of your xml. It starts with <data...>, and
#     ends with </data>. Be sure to copy everything between these <data...> and </data>, 
#     including these tags themselves. 
#
#3. From the command line, run the following command (in the directory where you saved 
#     the newly created file):
#       python uncertainty_transform.py DATAFILE P_THAT_?_IS_ONE P_THAT_ZERO_IS_ONE
#   Here, DATAFILE is the name of your file;
#     P_THAT_?_IS_ONE is the desired probability that a ? in your data was actually 1;
#     P_THAT_ZERO_IS_ONE is the desired probability that a 0 in your data was actually 1.
#
#4. Open the file DATAFILE, and replace the data node of your original BEAST 2 xml with a 
#     copy of its content, excluding the starting line opening with <?xml...
#
#
#Q: why doesn't the script transform the BEAST xml directly? 
#
#A: There are two reasons. First, BEAST xml files generated by BEAUti have specific 
#     formatting. Changing that formatting will make it much more difficult to edit the 
#     xml script by hand. But recreating the old formatting within this script would 
#     involve a lot of work, and anyway might lead to errors with some inputs, given the 
#     complexity of BEAST xml scripts, and the multitude of options that could be present 
#     there. The second reason is the hope is that by replacing the data section of your 
#     BEAST 2 xml yourself, you might have a better overall sense of what exactly 
#     changed. Should anything go wrong, which is not uncommon in the world of complex 
#     scripts such as BEAST 2 xmls, it will be easier for you to identify the problem. 
##########################################################################################
#Author: Igor Yanovich
##########################################################################################

import sys
import xmltodict

old_file = sys.argv[1]
transform_for_qu = str(1 - float(sys.argv[2])) + ', ' +  sys.argv[2] + '; '
transform_for_0 = str(1 - float(sys.argv[3])) + ', ' +  sys.argv[3] + '; '

print "Script uncertainty_transform.py has started"
print "Input data file:", sys.argv[1]
print "Probabilities to substitute for ?s:", transform_for_qu
print "Probabilities to substitute for 0s:", transform_for_0
print ""
print "Citation for this script:"
print ""
print "Yanovich, Igor. (2018) The effect of dictionary omissions"
print "  on phylogenies computationally inferred from lexical data."
print "Language Dynamics and Change."

with open(sys.argv[1]) as fd:
    myxml = xmltodict.parse(fd.read())
    
seqs = myxml['data']['sequence']
new_seqs = seqs

for index, seq in enumerate(seqs):
	data = seq['@value']
	new_data = ''
	for char in data:
		if char == '1':
			new_data += '0.0, 1.0; '
		elif char == '0':
			new_data += transform_for_0
		elif char == '?':
			new_data += transform_for_qu
		else:
			print "WARNING: In sequence", str(index), "an unexpected symbol:", char
			new_data += char
	new_seqs[index]['@value'] = new_data
	new_seqs[index]['@uncertain'] = 'true'
	
	
myxml['data']['sequence'] = new_seqs

with open(sys.argv[1] + '.uncertain', 'w') as newfile:
	xmltodict.unparse(myxml, output=newfile, pretty='True')			
